package Atlas::DB::DemographicsImport;

use strict;
use warnings;
use Carp qw(croak carp);

use DBI;
use Log::Log4perl qw(get_logger :levels);
use Data::Dumper;
use YAML qw(LoadFile);
use Readonly;
use List::Compare;
use Text::Trim;

Readonly my @APPOINTMENT_STRING_FIELDS => (
    'peoplesoft_department_code',
    'apso_department_code',
    'peoplesoft_pay_code',
    'appointment_status',
    [
        'appointment_time',
        'employee_type',
        'job_duration',
    ],
    'location',
    'hours_per_week',
    'appointment_type_code',
    'department_description',
    'job_title',
    'efs_dept_id_code',
    'rrc',
    'primary_appointment_flag',
    'zdept_department_code',
);

Readonly my @DB_FIELDS => (
    map { ref $_ eq ref [] ? @$_ : $_ }
    @APPOINTMENT_STRING_FIELDS
);

sub new {
    my $class = shift;
    my %args  = (
        normalization_lookup_table => {
            umnCampusMail     => \&normalize_address,
            homePostalAddress => \&normalize_address,
            umnOfficeAddress1 => \&normalize_address,
            psHomeAddress     => \&normalize_address,
            psMailAddress     => \&normalize_address,
            umnBirthDate      => \&normalize_date,
        },
        @_,
    );

    bless \%args, $class;
}

sub init_log {
    my $self = shift;
    my %args = (
        @_,
    );

    if ((exists $self->{'log4perl-config-file'}) && (-f $self->{'log4perl-config-file'})) {
        Log::Log4perl::init_and_watch($self->{'log4perl-config-file'}, $self->{'log4perl-config-file-watch-time'});
    }
    else {
        my $logger = get_logger(__PACKAGE__);
        $logger->level($INFO);

        my $appender = Log::Log4perl::Appender->new(
            'Log::Dispatch::Syslog',
        );
        $logger->add_appender($appender);

        my $layout = Log::Log4perl::Layout::PatternLayout->new("%d %p> %F{1}:%L %M - %m%n");
        $appender->layout($layout);
    }
}

sub init_map {
    my $self = shift;
    my %args = (
        @_,
    );

    my $logger = get_logger(__PACKAGE__);

    if (-f $self->{'map-file'}) {
        my $map_file = LoadFile($self->{'map-file'});
        if (defined $map_file) {
            $logger->debug('Mapping: ', { filter => \&Data::Dumper::Dumper, value => $map_file });
            $self->{mapping} = $map_file;
        }
        else {
            $logger->logcroak('Demographics mapping is not defined.');
        }
    }
    else {
        $logger->logcroak("File $self->{'map-file'} does not exist.");
    }

}

sub log_importer_beginning {
    my $self = shift;
    my %args = (
        @_,
    );

    my $logger = get_logger(__PACKAGE__);

    $logger->info('Demographics importer beginning.');
}

sub log_importer_finished {
    my $self = shift;
    my %args = (
        @_,
    );

    my $logger = get_logger(__PACKAGE__);

    $logger->info('Demographics importer finished.');
}

sub init_db {
    my $self = shift;
    my %args = (
        shall_die => 0,
        @_,
    );

    my $logger = get_logger(__PACKAGE__);

    # Set up some constants. They are somewhat arbitrary.
    Readonly my $PASSWORD                     => 'x';     # From man 5 passwd: If the password field is a lower-case "x", then the encrypted password is actually stored in the shadow(5) file instead;
    Readonly my $ENCRYPTED_PASSWORD           => '*';     # This seems to mean that there is no password hash
    Readonly my $DATE_OF_LAST_PASSWORD_CHANGE => 14856;   # A meaningless constant.
    Readonly my $MINIMUM_PASSWORD_AGE         => 0;       # We don't need to enforce a minimum password age.
    Readonly my $MAXIMUM_PASSWORD_AGE         => 99999;   # Make the password last as long as we can.
    Readonly my $PASSWORD_WARNING_PERIOD      => 7;       # One week warning period.
    Readonly my $PASSWORD_INACTIVITY_PERIOD   => 99999;   # No password inactivity period.
    Readonly my $ACCOUNT_EXPIRATION_DATE      => 99999;   # Don't expire according to this table.
    Readonly my $RESERVED_FIELD               => 0;       # Unused at this point. 0 is as good a value as anything.


    my $insert_query_parameters = [
        { field_name => 'username',                     value => '?',                           quote_value => 0, },
        { field_name => 'password',                     value => $PASSWORD,                     quote_value => 1, },
        { field_name => 'encrypted_password',           value => $ENCRYPTED_PASSWORD,           quote_value => 1, },
        { field_name => 'date_of_last_password_change', value => $DATE_OF_LAST_PASSWORD_CHANGE, quote_value => 0, },
        { field_name => 'minimum_password_age',         value => $MINIMUM_PASSWORD_AGE,         quote_value => 0, },
        { field_name => 'maximum_password_age',         value => $MAXIMUM_PASSWORD_AGE,         quote_value => 0, },
        { field_name => 'password_warning_period',      value => $PASSWORD_WARNING_PERIOD,      quote_value => 0, },
        { field_name => 'password_inactivity_period',   value => $PASSWORD_INACTIVITY_PERIOD,   quote_value => 0, },
        { field_name => 'account_expiration_date',      value => $ACCOUNT_EXPIRATION_DATE,      quote_value => 0, },
        { field_name => 'reserved_field',               value => $RESERVED_FIELD,               quote_value => 0, },
    ];

    my $update_query_parameters = [];

    while (my ($legacy_field, $atlas_field) = each %{$self->{mapping}{fields}{legacy_mapping}}) {
        $logger->debug("Pushing mapping legacy, $legacy_field, to Atlas field, $atlas_field, onto array.");
        my $push_item = { field_name => $atlas_field, value => '?', quote_value => 0, };
        push @$update_query_parameters,  $push_item;
        push @$insert_query_parameters, $push_item;
    }

    my $insert_sql
        = 'INSERT INTO users ('
        . join(', ', map { $_->{field_name} } @$insert_query_parameters)
        . ') VALUES ('
        . join(', ', map { ($_->{quote_value} ? q{'} : '').$_->{value}.($_->{quote_value} ? q{'} : '') } @$insert_query_parameters)
        . ')'
    ;
    $logger->debug("Created an insert SQL statement of $insert_sql");

    my $update_sql
        = 'UPDATE users SET '
        . join(', ', map { "$_->{field_name} = ?" } @$update_query_parameters)
        . ' WHERE username = ?'
    ;
    $logger->debug("Created an update SQL statement of $update_sql");

    my $done = -1;

    INITIALIZE_DATABASE:
    while ($done <= 0) {
        if ($done == 0) {
            sleep 5;
        }
        $done = 0;

        undef($self->{source_dbh});
        undef($self->{destination_dbh});

        my $source_database_directive      = ($self->{'source-database-driver'}      eq 'Pg') ? 'dbname' : 'database';
        my $destination_database_directive = ($self->{'destination-database-driver'} eq 'Pg') ? 'dbname' : 'database';

        my $source_dsn      = "DBI:$self->{'source-database-driver'}:$source_database_directive=$self->{'source-database-name'};host=$self->{'source-database-host'}";
        my $destination_dsn = "DBI:$self->{'destination-database-driver'}:$destination_database_directive=$self->{'destination-database-name'};host=$self->{'destination-database-host'}";

        $self->{'source_dbh'}      = DBI->connect($source_dsn,      $self->{'source-database-user'},      $self->{'source-database-pass'},      { RaiseError => 0, AutoCommit => 1 });
        $self->{'destination_dbh'} = DBI->connect($destination_dsn, $self->{'destination-database-user'}, $self->{'destination-database-pass'}, { RaiseError => 1, AutoCommit => 1 });

        # Verify connection to source DB
        if (! $self->{source_dbh}) {
            my $err = "Could not connect to $source_dsn. (".$DBI::errstr.")";
            $logger->error($err);
            croak $err if ($args{shall_die});
            if ($DBI::err == 1 || $DBI::err == 7) {
                next INITIALIZE_DATABASE;
            }
            else {
                $logger->logcroak('DBI died. We would normally restart, but we need the code. DBI err code: '.$DBI::err);
            }
        }

        # Verify connection to destination DB
        if (! $self->{destination_dbh}) {
            my $err = "Could not connect to $destination_dsn. (".$DBI::errstr.")";
            $logger->error($err);
            croak $err if ($args{shall_die});
            if ($DBI::err == 1 || $DBI::err == 7) {
                next INITIALIZE_DATABASE;
            }
            else {
                $logger->logcroak('DBI died. We would normally restart, but we need the code. DBI err code: '.$DBI::err);
            }
        }

        # Setup Statement Handles (sth). If those fail to be setup, "next" has the loop
        # forget about the DB connections and try again.
        next INITIALIZE_DATABASE if ! $self->init_sth(
            sql_key         => 'select_source_accounts',
            database_source => 'source',
            sth_sql         => $self->{'source-query'},
        );
        next INITIALIZE_DATABASE if ! $self->init_sth(
            sql_key         => 'select_destination_accounts',
            database_source => 'destination',
            sth_sql         => 'SELECT * FROM users',
        );
        next INITIALIZE_DATABASE if ! $self->init_sth(
            sql_key         => 'insert_into_destination',
            database_source => 'destination',
            sth_sql         => $insert_sql,
        );
        next INITIALIZE_DATABASE if ! $self->init_sth(
            sql_key         => 'update_destination_user',
            database_source => 'destination',
            sth_sql         => $update_sql,
        );
        next INITIALIZE_DATABASE if ! $self->init_sth(
            sql_key         => 'select_source_suppression_fields',
            database_source => 'source',
            sth_sql         => $self->{'source-suppressed-fields-query'},
        );
        next INITIALIZE_DATABASE if ! $self->init_sth(
            sql_key         => 'select_source_appointments',
            database_source => 'source',
            sth_sql         => $self->{'source-appointments-query'},
        );
        next INITIALIZE_DATABASE if ! $self->init_sth(
            sql_key         => 'select_destination_suppression_fields',
            database_source => 'destination',
            sth_sql         => 'SELECT * FROM suppressed_demographics',
        );
        next INITIALIZE_DATABASE if ! $self->init_sth(
            sql_key         => 'insert_suppression',
            database_source => 'destination',
            sth_sql         => 'INSERT INTO suppressed_demographics (university_id, suppressed_field) VALUES (?, ?)',
        );
        next INITIALIZE_DATABASE if ! $self->init_sth(
            sql_key         => 'delete_suppression',
            database_source => 'destination',
            sth_sql         => 'DELETE FROM suppressed_demographics WHERE university_id = ? AND suppressed_field = ?',
        );
        next INITIALIZE_DATABASE if ! $self->init_sth(
            sql_key         => 'select_destination_appointments',
            database_source => 'destination',
            sth_sql         => 'SELECT * FROM appointments',
        );
        next INITIALIZE_DATABASE if ! $self->init_sth(
            sql_key         => 'insert_appointment',
            database_source => 'destination',
            sth_sql         => 'INSERT INTO appointments (university_id, peoplesoft_department_code, apso_department_code, peoplesoft_pay_code, appointment_status, appointment_time, employee_type, job_duration, location, hours_per_week, appointment_type_code, department_description, job_title, efs_dept_id_code, rrc, primary_appointment_flag, zdept_department_code, verbatim) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)',
        );
        next INITIALIZE_DATABASE if ! $self->init_sth(
            sql_key         => 'delete_appointment',
            database_source => 'destination',
            sth_sql         => 'DELETE FROM appointments WHERE university_id = ? AND peoplesoft_department_code = ? AND apso_department_code = ? AND peoplesoft_pay_code = ? AND appointment_status = ? AND appointment_time = ? AND employee_type = ? AND job_duration = ? AND location = ? AND hours_per_week = ? AND appointment_type_code = ? AND department_description = ? AND job_title = ? AND efs_dept_id_code = ? AND rrc = ? AND primary_appointment_flag = ? AND zdept_department_code = ? AND verbatim = ?',
        );
        $done = 1;
    }
}

sub init_sth {
    my $self = shift;
    my %args = (
        sql_key         => undef,
        sth_sql         => undef,
        shall_die       => 0,
        database_source => undef, # source || destination
        @_,
    );
    my $db_src = $args{'database_source'};

    my $logger = get_logger(__PACKAGE__);

    my $sth_key = $args{sql_key}.'_sth';
    undef($self->{$sth_key});

    $self->{$args{sql_key}} = $args{sth_sql};

    $self->{$sth_key} = $self->{$db_src . '_dbh'}->prepare($args{sth_sql});
    if (! $self->{$sth_key}) {
        my $err = "Database error: ".$self->{$db_src . 'dbh'}->errstr." (while preparing '$args{sth_sql}')";
        $logger->error($err);
        croak $err if ($args{shall_die});
        if ($DBI::err == 1 || $DBI::err == 7) {
            return 0; # Return of 0 means to try again.
        }
        else {
            $logger->logcroak("DBI died. We would normally restart, but we need the code. DBI err code: ".$DBI::err);
        }
    }

    return 1; # Success
}

=head2
get_source_users - Get array of hashes of source users.
=cut
sub get_source_users {
    my $self = shift;
    my %args = (
        @_,
    );

    my $sql = 'select_source_accounts';
    my $sth = $sql.'_sth';

    my $source_users = undef;

    my $query_log_statement = "Query of [".$self->{$sql}."]";# with parameters of (".join(', ', @params).")";
    my $logger = get_logger(__PACKAGE__);

    my $done = -1;

    GET_SOURCE_USERS:
    while ($done <= 0) {
        if ($done == 0) {
            sleep 5;
        }
        $done = 0;
        if ($self->{$sth}->execute) {
            $source_users = $self->{$sth}->fetchall_arrayref({});
            if ($self->{$sth}->err) {
                my $err
                    = 'Database error: '.$self->{$sth}->errstr
                    . ' and database error code of '.$self->{$sth}->err
                    . " (while fetchall_hashref-ing '$query_log_statement')"
                ;
                if ($self->{$sth}->err == 1 || $self->{$sth}->err == 7) {
                    $err .= " However, found known error code and reinitializing database.";
                    $logger->warn($err);
                    $self->init_db;
                    next GET_SOURCE_USERS;
                }
                else {
                    $logger->error($err);
                }
            }
        }
        else {
            my $err
                = 'Database error: '.$self->{$sth}->errstr
                . ' and database error code of '.$self->{$sth}->err
                . " (while executing '$query_log_statement')"
            ;
            if ($self->{$sth}->err == 1 || $self->{$sth}->err == 7) {
                $err .= " However, found known error code and reinitializing database.";
                $logger->warn($err);
                $self->init_db;
                next GET_SOURCE_USERS;
            }
            else {
                $logger->fatal($err);
            }
        }
        $done = 1;
    }

    $self->{source_users} = $source_users;

    return;
}

=head2
get_destination_users - Get array of hashes of destionation users.
=cut

sub get_destination_users {
    my $self = shift;
    my %args = (
        @_,
    );

    Readonly my $DESTINATION_USER_KEY => 'username';

    my $sql = 'select_destination_accounts';
    my $sth = $sql.'_sth';

    my $destination_users = undef;

    my $query_log_statement = "Query of [".$self->{$sql}."]";# with parameters of (".join(', ', @params).")";
    my $logger = get_logger(__PACKAGE__);
    my $done = -1;

    GET_DESTINATION_USERS:
    while ($done <= 0) {
        if ($done == 0) {
            sleep 5;
        }
        $done = 0;

        if ($self->{$sth}->execute) {
            $destination_users = $self->{$sth}->fetchall_hashref($DESTINATION_USER_KEY);
            if ($self->{$sth}->err) {
                my $err
                    = 'Database error: '.$self->{$sth}->errstr
                    . ' and database error code of '.$self->{$sth}->err
                    . " (while fetchall_arrayref-ing '$query_log_statement')"
                ;
                if ($self->{$sth}->err == 1 || $self->{$sth}->err == 7) {
                    $err .= " However, found known error code and reinitializing database.";
                    $logger->warn($err);
                    $self->init_db;
                    next GET_DESTINATION_USERS;
                }
                else {
                    $logger->error($err);
                }
            }
        }
        else {
            my $err = 'Database error: '.$self->{$sth}->errstr
                    . ' and database error code of '.$self->{$sth}->err
                    . " (while executing '$query_log_statement')"
            ;
            if ($self->{$sth}->err == 1 || $self->{$sth}->err == 7) {
                $err .= " However, found known error code and reinitializing database.";
                $logger->warn($err);
                $self->init_db;
                next GET_DESTINATION_USERS;
            }
            else {
                $logger->fatal($err);
            }
        }
        $done = 1;
    }

    $self->{destination_users} = $destination_users;

    return;
}

sub normalize_source_data {
    my $self = shift;
    my %args = (
        source_record => undef,
        @_,
    );
    my $source_record = $args{source_record};

    my $logger = get_logger(__PACKAGE__);

    $logger->debug('Normalizing source record. ', { filter => \&Data::Dumper::Dumper, value => $source_record });

    NORMALIZE_LEGACY_DATA:
    for my $legacy_field (keys %$source_record) {
        if (! exists $self->{normalization_lookup_table}{$legacy_field}) {
            $logger->debug("Skipping field $legacy_field due to no lookup entry for normalization.");
            next NORMALIZE_LEGACY_DATA;
        }
        my $value = $source_record->{$legacy_field};

        $logger->debug(
            "Before normalizing source record field: $legacy_field = ",
            (defined $value ? $value : 'undefined'),
            '.',
        );
        $source_record->{$legacy_field} = $self->{normalization_lookup_table}{$legacy_field}(
            data => $value,
        );
        $logger->debug(
            "After normalizing source record field: $legacy_field = ",
            (defined $source_record->{$legacy_field} ? $source_record->{$legacy_field} : 'undefined'),
            '.',
        );
    }
}

sub normalize_address {
    my %args = (
        data => undef,
        @_,
    );

    my $data = $args{data};

    trim($data); # Remove leading and trailing space.

    Readonly my $LINE_SEPARATOR => qr{
        \s*    # optional whitespace
        \$     # a literal dollar sign
        \s*    # optional whitespace
    }xms;

    # Dollar signs are new line separators. So split on those and clean up any
    # whitespace surrounding the delimiters.
    my @address_lines = split $LINE_SEPARATOR, $data;

    # Correct continuation lines.
    # These are lines that are split between two lines. The continuation line
    # will begin with leading periods '....'.

    Readonly my $CONTINUATION_LINE_REGEX => qr{
        \A                            # Beginning of string
        (?<continuation_characters>
            \.{3,}                    # Literal dot (at least 3 of them.)
        )
    }xms;

    my $i = -1;

    LINE:
    for my $line (@address_lines) {
        $i++;
        if ($i == 0) {
            # There is no previous line to glue this line to.
            # No sense in even trying to "correct" it.
            next LINE;
        }

        if ($line =~ $CONTINUATION_LINE_REGEX) {
            # If this line is a continuation line, then
            # replace the continuation characters with a single space
            # so that we can append this line to the previous line.
            $line =~ s{$CONTINUATION_LINE_REGEX}{ };

            # Remove this continuation line.
            splice @address_lines, $i, 1;

            # And add it to the previous line.
            $address_lines[$i - 1] .= $line;
        }
    }

    # Put everything back together for the database
    $data = join "\n", @address_lines;

    return $data;
}

sub normalize_date {
    my %args = (
        data => undef,
        @_,
    );

    my $data = $args{data};

    my $logger = get_logger(__PACKAGE__);

    # undef values are okay, we'll just return it
    # and the DB can store it as NULL
    if (! defined $data) {
        return $data;
    }

    # I consider whitespace strings to be
    # NULL in the database. Return undef to set that.
    if ($data =~ /\A\s*\z/) {
        return undef;
    }

    my $regex = qr{
        \A
        \s*
        (?<year>\d{4})
        (?<separator>[-/]?)   # "separator" will be '-', '/', or '' (empty string)
        (?<month>\d{2})
        (?P=separator)?       # Use backreference to earlier "separator"
        (?<day>\d{2})
        \s*
        \z
    }xms;

    if ($data =~ $regex) {
        return "$+{year}-$+{month}-$+{day}";
    }
    else {
        # If the data didn't match our regex, then just
        # return the data and let the DB figure it out
        $logger->error("Data >>$data<< did not match date parser regex.");
        return $data;
    }
}

sub check_consistency {
    my $self = shift;
    my %args = (
        @_,
    );

    my $logger = get_logger(__PACKAGE__);

    my $users_processed_count = 0;
    my $source_users_count = @{$self->{source_users}};

    PROCESS_USER:
    for my $source_record (@{$self->{source_users}}) {
        $users_processed_count++;

        $self->normalize_source_data(
            source_record => $source_record,
        );
        my $source_username = $source_record->{umdemail}; # Convenience assignment. Would rather work with a scalar than a hashref.

        if ($source_username !~ m{
            \A     # beginning of string
            [\w-]+ # username must have only alphanumeric characters or a dash
            \z     # end of string
        }xms) {
            $logger->logcarp("Warning! Source user >>$source_username<< did not match the safety check, skipping. Details:", { filter => \&Data::Dumper::Dumper, value => $source_record });
            next PROCESS_USER;
        }

        if (($users_processed_count % 1000) == 1) {
            if (! $self->{quiet}) {
                my $percent_complete = sprintf("%.3f", ($users_processed_count / $source_users_count * 100));
                print "$percent_complete% Complete\n";
            }
        }

        $logger->debug("Source record: ", { filter => \&Data::Dumper::Dumper, value => $source_record });
        $logger->debug("Destination record for username $source_username ", { filter => \&Data::Dumper::Dumper, value => $self->{destination_users}{$source_username} });

        $self->normalize_source_data(
            source_record => $source_record,
        );

        if (exists $self->{destination_users}{$source_username}) {
            $logger->debug("Checking username $source_username fields in source and destination to determine if we need to update.");

            my @source_data;
            my @destination_data;

            # Iterate over the field mappings to get the data from both the source and destination records
            while (my ($legacy_field, $atlas_field) = each %{$self->{mapping}{fields}{legacy_mapping}}) {
                push @source_data, $source_record->{$legacy_field};
                push @destination_data, $self->{destination_users}{$source_username}{$atlas_field};
            }

            # Make a list compare object.
            my $list_compare = List::Compare->new('--unsorted', \@source_data, \@destination_data);

            # Test the size of the intersection between source and destination with size of the fields we are comparing
            if (scalar($list_compare->get_intersection) != scalar(keys %{$self->{mapping}{fields}{legacy_mapping}})) {;
                # Then we can update.
                $logger->debug("Updating $source_username in destination database.");
                $self->update_user(
                    source_record => $source_record,
                );
            }
        }
        else {
            $logger->debug("User $source_username does not exist in destination database.");
            $logger->info("Inserting $source_username into destination database.");
            $self->insert_user(
                source_record => $source_record,
            );
        }
    }
}

=head2
get_source_suppression_data - Get hash of hashes of source suppression data.
=cut
sub get_source_suppression_data {
    my $self = shift;
    my %args = (
        @_,
    );

    my $sql = 'select_source_suppression_fields';
    my $sth = $sql.'_sth';

    my $source_suppression_data = undef;

    my $query_log_statement = "Query of [".$self->{$sql}."]";
    my $logger = get_logger(__PACKAGE__);

    my $done = -1;

    GET_SOURCE_SUPPRESSION:
    while ($done <= 0) {
        if ($done == 0) {
            sleep 5;
        }
        $done = 0;
        if ($self->{$sth}->execute) {
            $source_suppression_data = $self->{$sth}->fetchall_arrayref({});
            if ($self->{$sth}->err) {
                my $err
                    = 'Database error: '.$self->{$sth}->errstr
                    . ' and database error code of '.$self->{$sth}->err
                    . " (while fetchall_arrayref-ing '$query_log_statement')"
                ;
                if ($self->{$sth}->err == 1 || $self->{$sth}->err == 7) {
                    $err .= " However, found known error code and reinitializing database.";
                    $logger->warn($err);
                    $self->init_db;
                    next GET_SOURCE_SUPPRESSION;
                }
                else {
                    $logger->error($err);
                }
            }
        }
        else {
            my $err
                = 'Database error: '.$self->{$sth}->errstr
                . ' and database error code of '.$self->{$sth}->err
                . " (while executing '$query_log_statement')"
            ;
            if ($self->{$sth}->err == 1 || $self->{$sth}->err == 7) {
                $err .= " However, found known error code and reinitializing database.";
                $logger->warn($err);
                $self->init_db;
                next GET_SOURCE_SUPPRESSION;
            }
            else {
                $logger->fatal($err);
            }
        }
        $done = 1;
    }

    if (defined $source_suppression_data) {
        my %suppression_data = ();

        # Convert the array of hashes data (from the database) into a hash of hashes data
        # structure.
        for my $suppression_entry (@$source_suppression_data) {
            my $university_id    = trim $suppression_entry->{umnid};

            # The legacy system will use the term 'entry' to mean suppress
            # all fields. We want to use the following, "__all_fields__"
            my $legacy_suppression_field = trim $suppression_entry->{umnSuppress};
            my $suppressed_field = '__all_fields__';

            # if the legacy suppression field is not "entry", then
            # we need to grab the correct mapping to our current standard.
            if ($legacy_suppression_field ne 'entry') {
                $suppressed_field = $self->{mapping}{fields}{legacy_mapping}{$legacy_suppression_field};
            }
            $suppression_data{$university_id}{$suppressed_field} = 1;
        }
        $self->{source_suppression_data} = \%suppression_data;
    }
    else {
        $self->{source_suppression_data} = undef;
    }

    return;
}

=head2
get_destination_suppression_data - Get hash of hashes of destination suppression data.
=cut
sub get_destination_suppression_data {
    my $self = shift;
    my %args = (
        @_,
    );

    my $sql = 'select_destination_suppression_fields';
    my $sth = $sql.'_sth';

    my $destination_suppression_data = undef;

    my $query_log_statement = "Query of [".$self->{$sql}."]";
    my $logger = get_logger(__PACKAGE__);

    my $done = -1;

    GET_DESTINATION_SUPPRESSION:
    while ($done <= 0) {
        if ($done == 0) {
            sleep 5;
        }
        $done = 0;
        if ($self->{$sth}->execute) {
            $destination_suppression_data = $self->{$sth}->fetchall_arrayref({});
            if ($self->{$sth}->err) {
                my $err
                    = 'Database error: '.$self->{$sth}->errstr
                    . ' and database error code of '.$self->{$sth}->err
                    . " (while fetchall_arrayref-ing '$query_log_statement')"
                ;
                if ($self->{$sth}->err == 1 || $self->{$sth}->err == 7) {
                    $err .= " However, found known error code and reinitializing database.";
                    $logger->warn($err);
                    $self->init_db;
                    next GET_DESTINATION_SUPPRESSION;
                }
                else {
                    $logger->error($err);
                }
            }
        }
        else {
            my $err
                = 'Database error: '.$self->{$sth}->errstr
                . ' and database error code of '.$self->{$sth}->err
                . " (while executing '$query_log_statement')"
            ;
            if ($self->{$sth}->err == 1 || $self->{$sth}->err == 7) {
                $err .= " However, found known error code and reinitializing database.";
                $logger->warn($err);
                $self->init_db;
                next GET_DESTINATION_SUPPRESSION;
            }
            else {
                $logger->fatal($err);
            }
        }
        $done = 1;
    }

    if (defined $destination_suppression_data) {
        my %suppression_data = ();

        # Convert the array of hashes data (from the database) into a hash of hashes data
        # structure.
        for my $suppression_entry (@$destination_suppression_data) {
            $suppression_data{$suppression_entry->{university_id}}{$suppression_entry->{suppressed_field}} = 1;
        }
        $self->{destination_suppression_data} = \%suppression_data;
    }
    else {
        $self->{destination_suppression_data} = undef;
    }

    return;
}

sub check_suppression_consistency {
    my $self = shift;
    my %args = (
        @_,
    );

    my $logger = get_logger(__PACKAGE__);

    # Find out which suppressed fields we need to add to atlas.
    while (my ($university_id, $suppressed_fields) = each %{$self->{source_suppression_data}}) {
        SOURCE_SUPPRESSED_FIELD:
        for my $suppressed_field (keys %$suppressed_fields) {
            if (! exists $self->{destination_suppression_data}{$university_id}{$suppressed_field}) {
                $self->insert_suppression(
                    university_id    => $university_id,
                    suppressed_field => $suppressed_field,
                );
            }
        }
    }

    # And find out which ones we need to delete from atlas.
    while (my ($university_id, $suppressed_fields) = each %{$self->{destination_suppression_data}}) {
        DESTINATION_SUPPRESSED_FIELD:
        for my $suppressed_field (keys %$suppressed_fields) {
            if (! exists $self->{source_suppression_data}{$university_id}{$suppressed_field}) {
                $self->delete_suppression(
                    university_id    => $university_id,
                    suppressed_field => $suppressed_field,
                );
            }
        }
    }
}

sub insert_user {
    my $self = shift;
    my %args = (
        @_,
    );
    my $source_record = $args{source_record}; # Convenience

    my $logger = get_logger(__PACKAGE__);

    my $sql = 'insert_into_destination';
    my $sth = $sql.'_sth';

    my @params = (
        $source_record->{umdemail},
    );

    while (my ($legacy_field, $atlas_field) = each %{$self->{mapping}{fields}{legacy_mapping}}) {
        my $source_record_legacy_field = $source_record->{$legacy_field};
        if (! defined $source_record_legacy_field) {
            $source_record_legacy_field = '[undef]';
        }
        $logger->debug("Pushing value, $source_record_legacy_field, for legacy field, $legacy_field, to params array.");
        push @params, $source_record->{$legacy_field};
    }

    $logger->debug("params for insert_user sql:", { filter => \&Data::Dumper::Dumper, value => \@params });

    my $query_log_statement
        = "Query of [".$self->{$sql}."] with parameters of ("
        . join(
            ', ',
            map
            { defined $_ ? $_ : '[undef]' }
            @params
        )
        . ")"
    ;

    my $done = -1;

    INSERT_USER:
    while ($done <= 0) {
        if ($done == 0) {
            sleep 2;
        }
        $done = 0;
        my $rows;
        eval {$rows = $self->{$sth}->execute(@params);};
        if ($@) {
            warn "DB Error Params:\n" . Dumper(\@params);
            return;
        }
        if ($rows) {
            if ($rows eq '0E0') {
                $logger->error("Inserting into destination database, expecting insert to affect 1 row, but 0 rows affected. $query_log_statement");
            }
            elsif ($rows > 1) {
                $logger->error("Inserting into destination database, expecting insert to affect 1 row, but more than 1 row affected. $query_log_statement");
            }
        }
        else {
            my $err
                = 'Database error: '.$self->{$sth}->errstr
                . ' and database error code of '.$self->{$sth}->err
                . " (while executing '$query_log_statement')"
            ;
            if ($self->{$sth}->err == 1 || $self->{$sth}->err == 7) {
                $err .= ' However, found known error code and reinitializing database.';
                $logger->warn($err);
                $self->init_db;
                next INSERT_USER;
            }
            else {
                $logger->fatal($err);
            }
        }
        $done = 1;
    }
    return;
}

sub insert_suppression {
    my $self = shift;
    my %args = (
        university_id    => undef,
        suppressed_field => undef,
        @_,
    );

    my $logger = get_logger(__PACKAGE__);

    my $sql = 'insert_suppression';
    my $sth = $sql.'_sth';

    my @params = (
        $args{university_id},
        $args{suppressed_field},
    );

    $logger->debug("params for insert_suppression sql:", { filter => \&Data::Dumper::Dumper, value => \@params });

    my $query_log_statement = "Query of [".$self->{$sql}."] with parameters of (".join(', ', @params).")";

    my $done = -1;

    INSERT_SUPPRESSION:
    while ($done <= 0) {
        if ($done == 0) {
            sleep 2;
        }
        $done = 0;
        my $rows;
        eval {
            $rows = $self->{$sth}->execute(@params);
        };
        if ($@) {
            warn "DB Error Params:\n" . Dumper(\@params);
            return;
        }
        if ($rows) {
            if ($rows eq '0E0') {
                $logger->error("Inserting into destination database, expecting insert to affect 1 row, but 0 rows affected. $query_log_statement");
            }
            elsif ($rows > 1) {
                $logger->error("Inserting into destination database, expecting insert to affect 1 row, but more than 1 row affected. $query_log_statement");
            }
        }
        else {
            my $err
                = 'Database error: '.$self->{$sth}->errstr
                . ' and database error code of '.$self->{$sth}->err
                . " (while executing '$query_log_statement')"
            ;
            if ($self->{$sth}->err == 1 || $self->{$sth}->err == 7) {
                $err .= ' However, found known error code and reinitializing database.';
                $logger->warn($err);
                $self->init_db;
                next INSERT_SUPPRESSION;
            }
            else {
                $logger->fatal($err);
            }
        }
        $done = 1;
    }
    return;
}

sub delete_suppression {
    my $self = shift;
    my %args = (
        university_id    => undef,
        suppressed_field => undef,
        @_,
    );

    my $logger = get_logger(__PACKAGE__);

    my $sql = 'delete_suppression';
    my $sth = $sql.'_sth';

    my @params = (
        $args{university_id},
        $args{suppressed_field},
    );

    $logger->debug("params for delete sql:", { filter => \&Data::Dumper::Dumper, value => \@params });

    my $query_log_statement = "Query of [".$self->{$sql}."] with parameters of (".join(', ', @params).")";

    my $done = -1;

    DELETE_SUPPRESSION:
    while ($done <= 0) {
        if ($done == 0) {
            sleep 2;
        }
        $done = 0;
        my $rows;
        eval {
            $rows = $self->{$sth}->execute(@params);
        };
        if ($@) {
            warn "DB Error Params:\n" . Dumper(\@params);
            return;
        }
        if ($rows) {
            if ($rows eq '0E0') {
                $logger->error("Deleting from destination database, expecting delete to affect 1 row, but 0 rows affected. $query_log_statement");
            }
            elsif ($rows > 1) {
                $logger->error("Deleting from destination database, expecting delete to affect 1 row, but more than 1 row affected. $query_log_statement");
            }
        }
        else {
            my $err
                = 'Database error: '.$self->{$sth}->errstr
                . ' and database error code of '.$self->{$sth}->err
                . " (while executing '$query_log_statement')"
            ;
            if ($self->{$sth}->err == 1 || $self->{$sth}->err == 7) {
                $err .= ' However, found known error code and reinitializing database.';
                $logger->warn($err);
                $self->init_db;
                next DELETE_SUPPRESSION;
            }
            else {
                $logger->fatal($err);
            }
        }
        $done = 1;
    }
    return;
}

sub update_user {
    my $self = shift;
    my %args = (
        @_,
    );
    my $source_record = $args{source_record}; # Convenience

    my $logger = get_logger(__PACKAGE__);

    my $sql = 'update_destination_user';
    my $sth = $sql.'_sth';

    my @params;

    while (my ($legacy_field, $atlas_field) = each %{$self->{mapping}{fields}{legacy_mapping}}) {
        $logger->debug(
            'Pushing value, ',
            (defined $source_record->{$legacy_field} ? $source_record->{$legacy_field} : 'undefined'),
            ", for legacy field, $legacy_field, to params array.",
        );
        push @params, $source_record->{$legacy_field};
    }
    push @params, $source_record->{umdemail};

    $logger->debug('params for update_user sql:', { filter => \&Data::Dumper::Dumper, value => \@params });

    my $query_log_statement
        = 'Query of ['
        . $self->{$sql}
        . '] with parameters of ('
        . join(', ', map({ defined $_ ? $_ : 'undefined' } @params)) # Translate undefined values to 'undefinded' string
        . ')'
    ;

    my $done = -1;

    UPDATE_USER:
    while ($done <= 0) {
        if ($done == 0) {
            sleep 2;
        }
        $done = 0;
        my $rows;
        eval {$rows = $self->{$sth}->execute(@params);};
        if ($@) {
            warn "DB Error Params:\n" . Dumper(\@params);
            return;
        }
        if ($rows) {
            if ($rows eq '0E0') {
                $logger->error("Updating user, expecting insert to affect 1 row, but 0 rows affected. $query_log_statement");
            }
            elsif ($rows > 1) {
                $logger->error("Updating user, expecting insert to affect 1 row, but more than 1 row affected. $query_log_statement");
            }
        }
        else {
            my $err
                = 'Database error: '.$self->{$sth}->errstr
                . ' and database error code of '.$self->{$sth}->err
                . " (while executing '$query_log_statement')"
            ;
            if ($self->{$sth}->err == 1 || $self->{$sth}->err == 7) {
                $err .= " However, found known error code and reinitializing database.";
                $logger->warn($err);
                $self->init_db;
                next UPDATE_USER;
            }
            else {
                $logger->fatal($err);
            }
        }
        $done = 1;
    }
    return;
}

sub get_destination_appointments {
    my $self = shift;
    my %args = (
        @_,
    );

    my $sql = 'select_destination_appointments';
    my $sth = $sql.'_sth';

    my $destination_appointments_data = undef;

    my $query_log_statement = "Query of [".$self->{$sql}."]";
    my $logger = get_logger(__PACKAGE__);

    my $done = -1;

    GET_DESTINATION_APPOINTMENTS:
    while ($done <= 0) {
        if ($done == 0) {
            sleep 5;
        }
        $done = 0;
        if ($self->{$sth}->execute) {
            $destination_appointments_data = $self->{$sth}->fetchall_arrayref({});
            if ($self->{$sth}->err) {
                my $err
                    = 'Database error: '.$self->{$sth}->errstr
                    . ' and database error code of '.$self->{$sth}->err
                    . " (while fetchall_arrayref-ing '$query_log_statement')"
                ;
                if ($self->{$sth}->err == 1 || $self->{$sth}->err == 7) {
                    $err .= " However, found known error code and reinitializing database.";
                    $logger->warn($err);
                    $self->init_db;
                    next GET_DESTINATION_APPOINTMENTS;
                }
                else {
                    $logger->error($err);
                }
            }
        }
        else {
            my $err
                = 'Database error: '.$self->{$sth}->errstr
                . ' and database error code of '.$self->{$sth}->err
                . " (while executing '$query_log_statement')"
            ;
            if ($self->{$sth}->err == 1 || $self->{$sth}->err == 7) {
                $err .= " However, found known error code and reinitializing database.";
                $logger->warn($err);
                $self->init_db;
                next GET_DESTINATION_APPOINTMENTS;
            }
            else {
                $logger->fatal($err);
            }
        }
        $done = 1;
    }

    if (defined $destination_appointments_data) {
        my %appointments_data = ();

        # Convert the array of hashes data (from the database) into a hash of
        # arrays of hashes data structure.
        for my $appointment (@$destination_appointments_data) {
            # Don't store the university_id in the "appointment" hash.
            # 1. We already have the university_id as the key to the outermost hash.
            # 2. When converting the source appointment string to a hash, it won't
            #    have the university_id field, thus making it more difficult to compare
            #    the two hashes.
            my $university_id = delete $appointment->{university_id};

            # And be *sure* to delete metadata from the (destination)
            # appointment hash. If we don't comparisons with the source
            # appointment will fail.
            delete $appointment->{inserted_at};

            $logger->debug(
                'Adding appointment to destination datastructure ',
                "for university_id >>$university_id<<: ",
                {
                    filter => \&Data::Dumper::Dumper,
                    value  => $appointment,
                },
            );
            push @{$appointments_data{$university_id}}, $appointment;
        }
        $self->{destination_appointments_data} = \%appointments_data;
    }
    else {
        $self->{destination_appointments_data} = undef;
    }

    return;
}

sub process_source_appointments {
    my $self = shift;
    my %args = (
        @_,
    );

    my $sql = 'select_source_appointments';
    my $sth = $sql.'_sth';

    my $source_appointments_data = undef;

    my $query_log_statement = "Query of [".$self->{$sql}."]";
    my $logger = get_logger(__PACKAGE__);

    my $done = -1;

    GET_SOURCE_APPOINTMENTS:
    while ($done <= 0) {
        if ($done == 0) {
            sleep 5;
        }
        $done = 0;
        if ($self->{$sth}->execute) {
            $source_appointments_data = $self->{$sth}->fetchall_arrayref({});
            if ($self->{$sth}->err) {
                my $err
                    = 'Database error: '.$self->{$sth}->errstr
                    . ' and database error code of '.$self->{$sth}->err
                    . " (while fetchall_arrayref-ing '$query_log_statement')"
                ;
                if ($self->{$sth}->err == 1 || $self->{$sth}->err == 7) {
                    $err .= " However, found known error code and reinitializing database.";
                    $logger->warn($err);
                    $self->init_db;
                    next GET_SOURCE_APPOINTMENTS;
                }
                else {
                    $logger->error($err);
                }
            }
        }
        else {
            my $err
                = 'Database error: '.$self->{$sth}->errstr
                . ' and database error code of '.$self->{$sth}->err
                . " (while executing '$query_log_statement')"
            ;
            if ($self->{$sth}->err == 1 || $self->{$sth}->err == 7) {
                $err .= " However, found known error code and reinitializing database.";
                $logger->warn($err);
                $self->init_db;
                next GET_SOURCE_APPOINTMENTS;
            }
            else {
                $logger->fatal($err);
            }
        }
        $done = 1;
    }

    if (defined $source_appointments_data) {
        $logger->debug('Source appointments data is defined.');
        for my $appointment (@$source_appointments_data) {
            $logger->debug("Source appointment:", { filter => \&Data::Dumper::Dumper, value => $appointment, });
            my $trimmed_appointment_string = trim $appointment->{umnJobSummary};
            my $trimmed_university_id      = trim $appointment->{umnid};
            my $hash = $self->parse_appointment_string(
                appointment_string => $trimmed_appointment_string,
            );
            $hash->{verbatim} = $trimmed_appointment_string;

            if (
                ! $self->remove_appointment_from_destination(
                    university_id      => $trimmed_university_id,
                    source_appointment => $hash,
                )
            ) {
                # If we were unable to delete (or remove) the appointment
                # from the destination datastructure, then it must be one
                # that atlas does not have. Therefore we need to add it
                # to the queue of appointments to insert.
                push
                    @{$self->{new_appointments_to_add_to_destination}},
                    {
                        university_id => $trimmed_university_id,
                        appointment   => $hash,
                    },
                ;
            }
        }
    }

    return;
}

sub remove_appointment_from_destination {
    my $self = shift;
    my %args = (
        university_id      => undef,
        source_appointment => undef,
        @_,
    );
    my $logger = get_logger(__PACKAGE__);
    $logger->trace(
        'Arguments:',
        {
            filter => \&Data::Dumper::Dumper,
            value  => \%args,
        },
    );

    # convenience variables
    my $university_id      = $args{university_id};
    my $source_appointment = $args{source_appointment};


    if (defined $self->{destination_appointments_data}) {
        my $destination_appointments = $self->{destination_appointments_data}{$university_id};

        my $i = 0;
        APPOINTMENT:
        for my $destination_appointment (@$destination_appointments) {
            if (
                $self->are_appointments_equal(
                    appointment_1 => $destination_appointment,
                    appointment_2 => $source_appointment,
                )
            ) {
                splice @$destination_appointments, $i, 1;

                # We successfully removed the "source" appointment from the
                # destination datastructure.
                return 1;
            }
            $i++;
        }
    }
    return 0;
}

sub are_appointments_equal {
    my $self = shift;
    my %args = (
        appointment_1 => undef,
        appointment_2 => undef,
        @_,
    );
    my $logger = get_logger(__PACKAGE__);
    $logger->trace(
        'Arguments:',
        {
            filter => \&Data::Dumper::Dumper,
            value  => \%args,
        },
    );

    # convenience variables
    my $appointment_1 = $args{appointment_1};
    my $appointment_2 = $args{appointment_2};

    # First make sure that the hash sizes are equal.
    if (keys %$appointment_1 != keys %$appointment_2) {
        $logger->debug(
            "Appointments 1 ($appointment_1->{verbatim}) ",
            "and 2 ($appointment_2->{verbatim}) have different ",
            "number of keys. Returning false.",
        );
        return 0;
    }

    # Then check to see that each key => value pair are
    # the same across both hashes.
    while (my ($key_1, $value_1) = each %$appointment_1) {
        $logger->trace(
            "Appointment 1 key >>$key_1<< and value >>$value_1<<.",
        );
        if ($value_1 ne $appointment_2->{$key_1}) {
            $logger->debug(
                "Key >>$key_1<< appointment 1 value ($value_1) ",
                "and appointment 2 value ($appointment_2->{$key_1}) ",
                "are different. Returning false. ",
            );
            return 0;
        }
    }

    $logger->debug(
        'Did not encounter any failure conditions. Returning true.',
    );
    return 1;
}

sub parse_appointment_string {
    my $self = shift;
    my %args = (
        appointment_string => undef,
        @_,
    );
    my $logger = get_logger(__PACKAGE__);
    $logger->trace(
        'Arguments:',
        {
            filter => \&Data::Dumper::Dumper,
            value  => \%args,
        },
    );

    # convenience variable
    my $appointment_string = $args{appointment_string};
    $logger->debug(
        "appointment string: >>$appointment_string<<",
    );

    my @appointment_field_values = split /:/, $appointment_string;

    my $retval_hash = {};

    my $i = 0;
    for my $field_name (@APPOINTMENT_STRING_FIELDS) {
        # Look up the field value from the split string
        my $field_value
            = ($i <= $#appointment_field_values)  # Does the split string array even have an entry for this index "i"
            ? trim($appointment_field_values[$i]) # if so, then use that value (just in case, trim value)
            : q{}                                 # if not, then use the empty string
        ;

        if (ref $field_name eq ref []) {
            my @appointment_details_field_values = split //, $field_value;

            if (@appointment_details_field_values > @$field_name) {
                $logger->warn(
                    'Number of fields in source stream is greater than what ',
                    'the database is designed for. Source data:',
                    {
                        filter => \&Data::Dumper::Dumper,
                        value  => \@appointment_details_field_values,
                    },
                    ' and database fields:',
                    {
                        filter => \&Data::Dumper::Dumper,
                        value  => $field_name,
                    },
                );
            }

            my $j = 0;
            for my $appointment_details_field_name (@$field_name) {
                # Look up the field value from the split string
                $retval_hash->{$appointment_details_field_name}
                    = ($j <= $#appointment_details_field_values)  # Does the split string array even have an entry for this index "j"
                    ? trim($appointment_details_field_values[$j]) # if so, then use that value (just in case, trim value)
                    : q{}                                         # if not, then use the empty string
                ;
                $j++;
            }
        }
        else {
            $retval_hash->{$field_name} = $field_value;
        }

        $i++;
    }
    $logger->debug("Parsed appointment:", { filter => \&Data::Dumper::Dumper, value => $retval_hash, });

    return $retval_hash;
}

sub delete_remaining_destination_appointments {
    my $self = shift;
    my %args = (
        @_,
    );

    my $logger = get_logger(__PACKAGE__);

    UNIVERSITY_ID:
    while (my ($university_id, $appointments) = each %{$self->{destination_appointments_data}}) {
        for my $appointment (@$appointments) {
            $self->delete_appointment(
                university_id => $university_id,
                appointment   => $appointment,
            );
        }
    }
}

sub delete_appointment {
    my $self = shift;
    my %args = (
        university_id => undef,
        appointment   => undef,
        @_,
    );

    # convenience variables
    my $university_id = $args{university_id};
    my $appointment   = $args{appointment};

    my $logger = get_logger(__PACKAGE__);

    my $sql = 'delete_appointment';
    my $sth = $sql.'_sth';

    my @params = (
        $university_id,
        (
            map { $appointment->{$_} }
            @DB_FIELDS
        ),
        $appointment->{verbatim},
    );
    $logger->info("Deleting appointment>> $appointment->{verbatim}");

    $logger->debug("params for delete sql:", { filter => \&Data::Dumper::Dumper, value => \@params });

    my $query_log_statement = "Query of [".$self->{$sql}."] with parameters of (".join(', ', @params).")";

    my $done = -1;

    DELETE_APPOINTMENT:
    while ($done <= 0) {
        if ($done == 0) {
            sleep 2;
        }
        $done = 0;
        my $rows;
        eval {
            $rows = $self->{$sth}->execute(@params);
        };
        if ($@) {
            warn "DB Error Params:\n" . Dumper(\@params);
            return;
        }
        if ($rows) {
            if ($rows eq '0E0') {
                $logger->error("Deleting from destination database, expecting delete to affect 1 row, but 0 rows affected. $query_log_statement");
            }
            elsif ($rows > 1) {
                $logger->error("Deleting from destination database, expecting delete to affect 1 row, but more than 1 row affected. $query_log_statement");
            }
        }
        else {
            my $err
                = 'Database error: '.$self->{$sth}->errstr
                . ' and database error code of '.$self->{$sth}->err
                . " (while executing '$query_log_statement')"
            ;
            if ($self->{$sth}->err == 1 || $self->{$sth}->err == 7) {
                $err .= ' However, found known error code and reinitializing database.';
                $logger->warn($err);
                $self->init_db;
                next DELETE_APPOINTMENT;
            }
            else {
                $logger->fatal($err);
            }
        }
        $done = 1;
    }
    return;
}

sub insert_new_appointments {
    my $self = shift;
    my %args = (
        @_,
    );

    my $logger = get_logger(__PACKAGE__);

    for my $bundled_appointment (@{$self->{new_appointments_to_add_to_destination}}) {
        $self->insert_appointment(
            university_id => $bundled_appointment->{university_id},
            appointment   => $bundled_appointment->{appointment},
        );
    }
}

sub insert_appointment {
    my $self = shift;
    my %args = (
        university_id => undef,
        appointment   => undef,
        @_,
    );

    # convenience variables
    my $university_id = $args{university_id};
    my $appointment   = $args{appointment};

    my $logger = get_logger(__PACKAGE__);

    my $sql = 'insert_appointment';
    my $sth = $sql.'_sth';

    my @params = (
        $university_id,
        (
            map { $appointment->{$_} }
            @DB_FIELDS
        ),
        $appointment->{verbatim},
    );

    $logger->info("Insert appointment>> $appointment->{verbatim}");
    $logger->debug("params for insert_appointment sql:", { filter => \&Data::Dumper::Dumper, value => \@params });

    my $query_log_statement = "Query of [".$self->{$sql}."] with parameters of (".join(', ', @params).")";

    my $done = -1;

    INSERT_APPOINTMENT:
    while ($done <= 0) {
        if ($done == 0) {
            sleep 2;
        }
        $done = 0;
        my $rows;
        eval {
            $rows = $self->{$sth}->execute(@params);
        };
        if ($@) {
            warn "DB Error Params:\n" . Dumper(\@params);
            return;
        }
        if ($rows) {
            if ($rows eq '0E0') {
                $logger->error("Inserting into destination database, expecting insert to affect 1 row, but 0 rows affected. $query_log_statement");
            }
            elsif ($rows > 1) {
                $logger->error("Inserting into destination database, expecting insert to affect 1 row, but more than 1 row affected. $query_log_statement");
            }
        }
        else {
            my $err
                = 'Database error: '.$self->{$sth}->errstr
                . ' and database error code of '.$self->{$sth}->err
                . " (while executing '$query_log_statement')"
            ;
            if ($self->{$sth}->err == 1 || $self->{$sth}->err == 7) {
                $err .= ' However, found known error code and reinitializing database.';
                $logger->warn($err);
                $self->init_db;
                next INSERT_APPOINTMENT;
            }
            else {
                $logger->fatal($err);
            }
        }
        $done = 1;
    }
    return;
}

sub debug {
    my $self = shift;
    print "Source Users:\n";
    print Dumper($self->{'source_users'});
    print "Destination Users:\n";
    print Dumper($self->{'destination_users'});
    return;
}

1;
